/*
 *  cache.c
 *  mod_musicindex
 *
 *  $Id: cache.c,v 1.22 2003/07/16 01:35:45 varenet Exp $
 *
 *  Created by Thibaut VARENE on Fri Jul 04 2003.
 *  Copyright (c) 2003 Thibaut VARENE
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */

/**
 * @file 
 * Cache management subsystem.
 *
 * Any cache implementation (regular files, SQL, or whatever backend)
 * ought to provide the 4 mandatory interfaces:
 * short cache_check_dir(request_rec *, mu_config *, mu_cache_data *)
 * short cache_check_file(request_rec *, mu_config *, mu_cache_data *)
 * mu_ent *cache_read_file(request_rec *, mu_ent *, mu_config *, mu_cache_data *)
 * short cache_write_file(request_rec *, mu_ent *, mu_config *, mu_cache_data *)
 *
 * @author Thibaut Varene
 * @version $Revision: 1.22 $
 * @date 2003
 *
 * @warning Under heavy development, not very well tested!
 * @warning 'errno' seems not to be seen as volatile by the compiler, therefore we
 *	cannot do "if(mkdir(foo) && errno == EEXIST)" for instance.
 *
 * @bug We do rely on success of chdir() calls for the flat file implementation.
 *
 * @todo Thorough testing! (on various platform if possible)
 * @todo Use Apache API instead of standard API (ap_fopen, ap_popendir, etc...)?
 * @todo Write SQL bits.
 *
 * @test Shows amazing perfs on jukebox while requesting huge amount of data
 *	(recursive play on npyu for instance). The memory leak doesn't seem
 *	to come from here. Suspect something screwed with ap_palloc() in playlist.c.
 */

#include "cache.h"

#undef CACHE_SQL
#ifdef CACHE_SQL
/* A coder */
short cache_check_dir(request_rec * r, mu_config *conf, mu_cache_data *cachedata);
short cache_check_file(request_rec * r, mu_config *conf, mu_cache_data *cachedata);
mu_ent *cache_read_file(request_rec * r, mu_ent *head, mu_config *conf, mu_cache_data *cachedata);
short cache_write_file(request_rec * r, mu_ent *p, mu_config *conf, mu_cache_data *cachedata);

#else	/* Default cache system: filesystem hierarchy with regular text files */

/**
 * Handles error for the flat file cache subsystem.
 *
 * This function handles various errors depending on errno's value.
 * 
 * @param r Apache request_rec struct to handle log writings.
 * @param caller A string (eg. calling function name) used in messages sent.
 *
 * @todo Everything.
 */
static void error_handler(request_rec *r, const char *caller)
{
/* La ya bcp de boulot, vu que rien n'est fait...
Faut definir des codes d'erreur de retour eventuels (pour
EEXIST par ex...) et definir le fonctionnement selon le niveau d'erreur...
Bref, de quoi s'amuser. */

	ap_log_printf(r->server, "%s - error_handler! errno=%i", caller, errno);
	switch (errno) {
		case EPERM:
			/* The filesystem containing pathname does not support the creation of directories. */
		case EISDIR:
			/* pathname refers to a directory.  (This is  the  non-POSIX  value returned by Linux since 2.1.132.) */
		case EINVAL:
			/* mode  requested  creation of something other than a normal file, device special file or FIFO */
		case EEXIST:
			/* pathname already exists (not necessarily as a directory). */
		case EFAULT:
			/* pathname points outside your accessible address space. */
		case EACCES:
			/* The parent directory does not allow write permission to the  process,  or  one  of  the
			directories in pathname did not allow search (execute) permission. */
		case EMFILE:
			/* Too many file descriptors in use by process. */
		case ENFILE:
			/* Too many files are currently open in the system. */
		case ENAMETOOLONG:
			/* pathname was too long. */
		case ENOENT:
			/* A directory component in pathname does not exist or is a dangling symbolic link. */
		case ENOTDIR:
			/* A component used as a directory in pathname is not, in fact, a directory. */
		case ENOTEMPTY:
			/* pathname contains entries other than . and .. . */
		case ENOMEM:
			/* Insufficient kernel memory was available. */
		case EROFS:
			/* pathname refers to a file on a read-only filesystem. */
		case ELOOP:
			/* Too many symbolic links were encountered in resolving pathname. */
		case EIO:
			/* Am I/O error occured. */
		case ENOSPC:
			/* The device containing pathname has no room for the new directory.
			The new directory cannot be created because the user's disk quota is exhausted. */
		default:
			break;
		}
	return;
}

/**
 * Creates cache subdirectories.
 *
 * This subroutine takes care of creating all requested directories and
 * subdirectories if they don't already exist and if possible.
 * 
 * @warning If dirpath begins with a "/" the function will return immediately.
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param dirpath A string representing a path to create.
 *
 * @return 0 on succes, error code otherwise.
 */
static short cache_make_dir(request_rec *r, char *dirpath)
{
	short l = 0, m = 0;
	char *tempdir = NULL;
	
	do {	/* We build the path subdirs by subdirs, in a "mkdir -p" fashion */
		tempdir = ap_pstrndup(r->pool, dirpath, (m + (l = strcspn(dirpath + m, "/"))));
		m += l;
		
		if (!l)
			break;

		/* skipping (potentially multiple) slashes */
		while (dirpath[m] == '/')
			m++;
		
		if (mkdir(tempdir, S_IRWXU)) {
			if (errno == EEXIST);
			else
				goto error_out;
		}
	} while (1);

	return 0;

error_out:
	error_handler(r, __FUNCTION__);
	return CA_FATAL;
}

/**
 * Removes cache subdirectories.
 *
 * This subroutine takes care of removing any given directory and
 * its content (recursively) if any, and if possible.
 * 
 * @param r Apache request_rec struct to handle log writings.
 * @param cachedir A DIR stream corresponding to the directory to remove.
 * @param curdir A string representing the absolute path of the corresponding
 *	parent directory on the "original" filesystem.
 */
static void cache_remove_dir(request_rec *r, DIR *cachedir, char *curdir)
{
	DIR		*subdir = NULL;
	struct dirent	*cachedirent = NULL;
	struct stat	origdirstat;
	char 		*origdir = NULL;
	
	fchdir(dirfd(cachedir));			/* on se place dans le repertoire de cache. XXX Pas de test ici, tout est verifie avant (en principe). */
	while ((cachedirent = readdir(cachedir))) {	/* on parcours le repertoire */
		if (!(strcmp(cachedirent->d_name, ".")) || !(strcmp(cachedirent->d_name, "..")))	/* We'd rather avoid trying to remove the whole filesystem... */
			continue;
		
		if (unlink(cachedirent->d_name)) {	/* We try to remove any entry (actually we will only remove regular files) */
			if (errno == EISDIR) {		/* Unfortunately errno isn't seen as volatile, so we have to separate the if() */
				/* If it's a directory, we check that the "original" still exists.
				 * If not, we remove it recursively */
				origdir = ap_pstrcat(r->pool, curdir, "/", cachedirent->d_name, NULL);
				if (stat(origdir, &origdirstat)) {
					if (rmdir(cachedirent->d_name)) {	/* stat() sets errno. We have to split */
						if (errno == ENOTEMPTY) {			/* il est pas vide, bigre! */
							subdir = opendir(cachedirent->d_name);	/* on ouvre le vilain repertoire pour en supprimer le contenu */
							cache_remove_dir(r, subdir, origdir);	/* en rappelant recursivement la fonction sur son contenu. */
							closedir(subdir);			/* a noter que dans ce cas la il y a un test inutile, celui qui verifie si l'original existe tjrs. Mais bon. */
							fchdir(dirfd(cachedir));		/* on retourne au repertoire precedent */
							rmdir(cachedirent->d_name);		/* maintenant il est vide, et on peut pas avoir d'erreur vu les tests precedants */
						}
						else
							error_handler(r, __FUNCTION__);			/* Oops, on est tombe sur une merde */
					}
				}
			}
			else
				error_handler(r, __FUNCTION__);		/* Oops, on est tombe sur une merde, mais plus tot */
		}
	}
	
	return;
}

/**
 * Initializes flat file cache subsystem.
 *
 * Basically we do nothing more here than creating the root cache folder.
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param conf configuration structure in which the path to the cache root can be found.
 *
 * @return 0 on succes, error code otherwise.
 */
static short cache_init(request_rec * r, mu_config *conf)
{
	chdir("/");	/* let's pray this will never fail */
	if (cache_make_dir(r, conf->cache_path + 1))	/* since we've chdir'd, we send the path without the leading '/' */
		goto error_out;
	
	return 0;
	
error_out:
	error_handler(r, __FUNCTION__);
	return CA_FATAL;
}

/**
 * Checks if a directory already exists in the cache.
 *
 * This function is called when the callee wants to know whether a given
 * directory (as found in the cachedata struct) exists or not in the cache.
 * If the directory exists, the function returns successfully, otherwise
 * it tries to create it if possible and returns accordingly.
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param conf The config structure used to find out cache configuration.
 * @param cachedata A structure in which the full directory path can be found.
 *
 * @return 0 on succes, error code otherwise.
 */
short cache_check_dir(request_rec * r, mu_config *conf, mu_cache_data *cachedata)
{
	DIR		*cachedir = NULL;
	struct stat	cachedirstat, dirstat;
	
	if (!cachedata->name)
		return CA_MISSARG;
	
	/* Making sure the cache has been initialized, initialize it otherwise.
	 * Bear in mind we're chdir'd from now on. */
	if (chdir(conf->cache_path)) {	 		/* on va dans le rep de cache */
		if (errno == ENOENT) {			/* il n'existe pas, il est temps d'initialiser le cache */
			if (cache_init(r, conf))
				return CA_FATAL;	/* l'init du cache a chie, c'est mauvais */
			chdir(conf->cache_path);	/* maintenant qu'on l'a cree on peut enfin y aller. Considere sans echec */
		}
		else
			goto error_out;			/* un autre probleme, on degage */
	}
	
	/* Actually check for the directory in the cache, create it if needed.
	 * "+ 1" offset to suppress leading '/'. */
	if (!(cachedir = opendir(cachedata->name + 1))) {		/* on essaye d'ouvrir le repertoire concerne dans le cache (on supprime le leading "/" */
		if (errno == ENOENT) {					/* il n'existe pas mais on peut le creer (ca correspond a ENOENT, a verifier) */
			if (cache_make_dir(r, cachedata->name + 1))	/* on envoie le chemin prive du leading '/' */
				goto error_out;
		}
		else
			goto error_out;					/* un autre probleme, on degage */
	}
	else {	/* Checking for cache sanity. Has it expired for that folder ? If so, delete its content. */
		fstat(dirfd(cachedir), &cachedirstat);			/* recuperons les stats du repertoire. XXX On considere cet appel sans echec vu les tests qu'on a fait avant. */	
		stat(cachedata->name, &dirstat);			/* recuperons les stats du rep d'origine. XXX pas de test ici, a priori ya pas de raison qu'on puisse pas les recuperer */
		if (cachedirstat.st_mtime < dirstat.st_mtime)		/* si la date de modif du rep de cache est plus vieille que celle du rep original, alors qqc a ete ajoute ou retire ou ecrit */
			cache_remove_dir(r, cachedir, cachedata->name);	/* alors on le vide proprement */
		closedir(cachedir);					/* On en a fini avec le repertoire, on le referme */
	}

	return 0;

error_out:
	error_handler(r, __FUNCTION__);
	return CA_FATAL;
}

/**
 * Checks if a file already exists in the cache.
 *
 * This function is called when the callee wants to know whether a given
 * file (as found in the cachedata struct) exists or not in the cache.
 * If the file exists, the function returns successfully, providing the
 * FILE stream in cachedata->stream, otherwise it returns a specific error code.
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param conf The config structure used to find out cache configuration.
 * @param cachedata A structure in which the full file path can be found
 *	and in which the file stream will be set if needed.
 *
 * @return 0 on succes, CA_CREATE if the file doesn't exist but can be created,
 *	error code otherwise.
 */
short cache_check_file(request_rec * r, mu_config *conf, mu_cache_data *cachedata)
{
	if (!cachedata->name)
		return CA_MISSARG;
	
	/* Making sure the cache has been initialized, initialize it otherwise.
	 * Bear in mind we're chdir'd from now on. */
	if (chdir(conf->cache_path)) {		 	/* on va dans le rep de cache */
		if (errno == ENOENT) {			/* il n'existe pas, il est temps d'initialiser le cache */
			if (cache_init(r, conf))
				return CA_FATAL;	/* l'init du cache a chie, c'est mauvais */
			chdir(conf->cache_path);	/* maintenant qu'on l'a cree on peut enfin y aller. Considere sans echec */
		}
		else
			goto error_out;			/* un autre probleme, on degage */
	}

	/* Actually check for the file in the cache, open it if possible.
	 * "+ 1" offset to suppress leading '/'. */
	if (!(cachedata->stream = fopen(cachedata->name + 1, "r"))) {	/* on essaye d'ouvrir le fichier concerne en ro */
		if (errno == ENOENT)					/* il n'existe pas mais on peut a priori le creer (ca correspond a ENOENT, a verifier) */
			return CA_CREATE;
		else
			goto error_out;					/* un autre probleme, on degage */
	}
	
	/* The way we handle cache sanity in cache_check_dir() ensures that data is always up-to-date in cache files. */
	return 0;			/* le fichier existe et est ouvert, on peu sortir. */
	
error_out:
	error_handler(r, __FUNCTION__);
	return CA_FATAL;
}

/**
 * Reads cache file information and fills mu_ent entry accordingly.
 *
 * This function instanciates a new mu_ent entry and fills in its fields
 * with data gathered from the cache file.
 *
 * @bug Consummes 3*1k of RAM for each file read; we could probably shorten
 *	that to 3*256B, but the field can be up to 1k long according to id3.org
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param conf The config structure used to find out cache configuration.
 * @param head The current mu_ent pointer.
 * @param cachedata A structure in which the cache file stream can be found.
 *
 * @return A new mu_ent entry, or head if failure.
 */
mu_ent *cache_read_file(request_rec * r, mu_ent *head, mu_config *conf, mu_cache_data *cachedata)
{
	mu_ent		*p = head;
	short 		result = 0;

	p = new_ent(r->pool, head);
	p->album = ap_palloc(r->pool, MAX_STRING);	/* ugh..ly! */
	p->artist = ap_palloc(r->pool, MAX_STRING);
	p->title = ap_palloc(r->pool, MAX_STRING);
	
	result = fscanf(cachedata->stream, "album: %[^\n]\nartist: %[^\n]\n"
		"title: %[^\n]\ndate: %hu\ntrack: %hu\nposn: %hu\n"
		"length: %lu\nbitrate: %lu\nsize: %lu\nfiletype: %c\n",
		p->album, p->artist, p->title, &p->date, &p->track, &p->posn, &p->length,
		&p->bitrate, &p->size, &p->filetype);
	
	fclose(cachedata->stream);
	
	/* XXX ameliorer le test ici. if (result != NB_FIELDS) ? */
	if (!result)	/* fscanf() returns the number of input items assigned */
		return head;
	
	if (!strcmp(p->album, "(null)"))
		p->album = '\0';
	if (!strcmp(p->artist, "(null)"))
		p->artist = '\0';
	
	return p;
}

/**
 * Creates and writes cache file information.
 *
 * This function creates a new cache file (using cachedata->name), and
 * fills it with the data contained in the mu_ent p structure.
 *
 * @param r Apache request_rec struct to handle log writings.
 * @param conf The config structure used to find out cache configuration.
 * @param p A mu_ent struct containing the data to store.
 * @param cachedata A structure in which the cache file name can be found.
 *
 * @return 0 on success, error code otherwise.
 */
short cache_write_file(request_rec * r, mu_ent *p, mu_config *conf, mu_cache_data *cachedata)
{
	chdir(conf->cache_path);	/* Par securite on se re-chdir(). XXX Considere sans echec */
	
	if (!(cachedata->stream = fopen(cachedata->name + 1, "w+")))	/* on cree le fichier en le tronquant des fois qu'il existe deja (fichier pourri). */
		goto error_out;			/* decidement ca chie, on degage */

	fprintf(cachedata->stream, "album: %s\nartist: %s\ntitle: %s\ndate: %hu\n"
		"track: %hu\nposn: %hu\nlength: %lu\nbitrate: %lu\n"
		"size: %lu\nfiletype: %s\n",
		p->album, p->artist, p->title, p->date, p->track, p->posn, p->length,
		p->bitrate, p->size, &p->filetype);

	/* XXX check sur fprintf pour verifier que tout a ete ecrit? */
	fclose(cachedata->stream);
	
	return 0;

error_out:
	error_handler(r, __FUNCTION__);
	return CA_FATAL;
}
#endif
