/*
 *  mod_musicindex.c
 *  mod_musicindex
 *
 *  $Id: mod_musicindex.c,v 1.38 2003/07/16 01:35:45 varenet Exp $
 *
 *  Created by Regis BOUDIN on Sat Dec 28 2002.
 *  Copyright (c) 2003 Regis BOUDIN
 *  Copyright (c) 2003 Thibaut VARENE
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */


/**
 * @file 
 * Core file.
 *
 * This file is the core of the module. It contains Apache's mandatory stuff.
 *
 * @author Regis Boudin
 * @version $Revision: 1.38 $
 * @date 2003
 *
 * @warning Use it at your own risks!
 *
 * @todo MORE DOCUMENTATION OF THE CODE!!!
 * @todo custom playlist handling accross several directories.
 * @todo tag rewriting (eventually).
 * @todo Considering the previous item, we might also need a handler for
 * streaming, otherwise we won't be able to restrain file downloading
 * (Apache::MP3 doesn't do anything better there fwiw).
 * @todo eventually prepare the possibility to generate ices playlists.
 */

#include "mod_musicindex.h"
#include "playlist.h"
#include "config.h"
#include "html.h"
#include "inf.h"

/**
 * Creates a new musical entry (mp3, ogg or directory).
 *
 * @param pool Apache pool linked to the current request
 * @param head music entry structure which will follow this one in the list
 *
 * @return A new mu_ent
 */
struct mu_ent *new_ent(pool *pool, struct mu_ent *head)
{
	struct mu_ent *p = (struct mu_ent *) ap_pcalloc(pool, sizeof(struct mu_ent));
	memset(p, 0, sizeof(struct mu_ent));	
	p->next = head;
	return p;
}

/**
 * Sorts the list of musical entries.
 *
 * This is a fully home-made quicksort implementation. It places directory
 * entries first, then songs accordingly to the configuration.
 *
 * @todo Move with related functions (in inf.c)
 *
 * @param base The first entry of the portion to sort
 * @param end The last entry of the portion to sort
 * @param conf The configuration structure
 *
 * @return The first entry of the sorted list
 */
struct mu_ent *quicksort(struct mu_ent *base, struct mu_ent *end,
		      mu_config * conf) 
{
	struct mu_ent *prem;
	struct mu_ent *deux = base;
	/* The first entry will be used as reference */
	struct mu_ent *refe = base;

	/* If we have 0 or one entry, sorting is very easy */
	if ((base == end) || (base->next == end))
		return base;

	/* We will begin comparisons with the second entry of the list */
	prem = base->next;

	do {
		/* If the current entry is not correctly placed,
		move it before the reference */
		if (inf_global(prem, refe, conf) < 0) {
			deux->next = prem->next;
			prem->next = base;
			base = prem;
			prem = deux->next;
		/* Otherwise, just continue working with the next one */
		} else {
			deux = prem;
			prem = prem->next;
		}
	} while (prem != end);

	/* We now have two lists, A & B, separated by a value V,
	considering A < V < B . Just run the same quicksort on A and B */
	base = quicksort(base, refe, conf);
	refe->next = quicksort(refe->next, end, conf);

	return base;
}

int handle_musicindex(request_rec * r)
{
	char *ifile;
	/* Get the module configuration */
	mu_config *conf = (mu_config *)
		ap_get_module_config(r->per_dir_config, &musicindex_module);

	r->allowed |= (1 << M_GET);
	/* if the request is not a GET and the module is not active, let the dir
	to another module */
	if ((r->method_number != M_GET) || (!(conf->options & MI_ACTIVE)))
		return DECLINED;

	/* This part mostly comes from mod_autoindex. If the requested dir does
	not end with a '/', generate a new request with it */
	if (r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/') {
		if (r->args != NULL)
			ifile = ap_pstrcat(r->pool, ap_escape_uri(r->pool, r->uri),
		      "/", "?", r->args, NULL);
		else
			ifile = ap_pstrcat(r->pool, ap_escape_uri(r->pool, r->uri),
		      "/", NULL);

		ap_table_setn(r->headers_out, "Location",
		ap_construct_url(r->pool, ifile, r));
		return HTTP_MOVED_PERMANENTLY;
	}

	/* Depending on parameters from the HTTP request, call the adequate
	function to generate either a webpage or a m3u playlist */
	switch (treat_args(r, conf)) {
		case 'A':
			return musicindex_directory(r, conf);
		case 'P':
			return playlist_directory(r, conf);
		case 'L':
			return playlist_selected(r, conf);
		default:
			return 0;
	}

}

/**
 * Handler for requests on mp3 files.
 *
 * At the moment it can only forbid acces to files if neither streaming nor download is allowed.
 *
 * @param r Apache request_rec to get and send data
 *
 * @return Standard HTTP code (see httpd.h for more infos)
 */
int handle_mp3(request_rec * r)
{
	mu_config *conf = (mu_config *)
		ap_get_module_config(r->per_dir_config, &musicindex_module);

	if (!(conf->options&(MI_ALLOWSTREAM|MI_ALLOWDWNLD)))
		return HTTP_FORBIDDEN;
	return DECLINED;
}

/**
 * Handler for requests on ogg files.
 *
 * At the moment it can only forbid acces to files if neither streaming nor download is allowed.
 *
 * @param r Apache request_rec to get and send data
 *
 * @return Standard HTTP code (see httpd.h for more infos)
 */
int handle_ogg(request_rec *r)
{
	mu_config *conf = (mu_config *)
		ap_get_module_config(r->per_dir_config, &musicindex_module);

	if (!(conf->options&(MI_ALLOWSTREAM|MI_ALLOWDWNLD)))
		return HTTP_FORBIDDEN;
	return DECLINED;
}

const command_rec musicindex_cmds[] = {
	{"MusicSortOrder", sort_order, NULL, OR_INDEXES, RAW_ARGS,
		"can be : title album artist track disc length bitrate filetype"},
	{"MusicFields", set_fields, NULL, OR_INDEXES, RAW_ARGS,
		"can be : title album artist track disc length bitrate filetype"},
	{"MusicAllowStream", allow_stream, NULL, OR_INDEXES, NO_ARGS,
		"Enable streaming"},
	{"MusicAllowDownload", allow_download, NULL, OR_INDEXES, NO_ARGS,
		"Enable download"},
	{"MusicAllowSearch", allow_search, NULL, OR_INDEXES, NO_ARGS,
		"Enable searching (EXPERIMENTAL)"},
	{"MusicLister", music_lister, NULL, OR_INDEXES, NO_ARGS,
		"Enable the Musicindex maker"},
	{"MusicCachePath", set_cache_path, NULL, OR_INDEXES, TAKE1,
		"Set the cache absolute path, without trailing '/'."},
	{NULL}
};

const handler_rec musicindex_handlers[] = {
	{DIR_MAGIC_TYPE, handle_musicindex},
	{"audio/mpeg", handle_mp3},
	{"audio/x-ogg", handle_ogg},
	{NULL}
};

module MODULE_VAR_EXPORT musicindex_module = {
	STANDARD_MODULE_STUFF,
	NULL,				/**< initializer */
	create_musicindex_config,	/**< dir config creater */
	merge_musicindex_configs,	/**< dir merger */
	NULL,				/**< server config */
	NULL,				/**< merge server config */
	musicindex_cmds,		/**< command table */
	musicindex_handlers,		/**< handlers */
	NULL,				/**< filename translation */
	NULL,				/**< check_user_id */
	NULL,				/**< check auth */
	NULL,				/**< check access */
	NULL,				/**< type_checker */
	NULL,				/**< fixups */
	NULL,				/**< logger */
	NULL,				/**< header parser */
	NULL,				/**< child_init */
	NULL,				/**< child_exit */
	NULL				/**< post read-request */
};
